/* data_new.c */
/* Will Bridewell */

/* This code will read a file of space delimited values.
 * The first line should contain field names.
 * The remaining lines should contain *numeric* values.
 * The first field should always be t for time.
 * The first field should have a value for each observation point.
 * That is, each set of synchronous observations must be associated with a time.
 */

#include "data.h"
#include "model.h"
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <string.h>

/* the array where we will store field names */
char **field_names;

/* the number of fields in the file */
int nFields;

/* the table of values */
/* rows are the rows in the table */
/* columns are the columns in the table */
double **observed_values;

/* the number of total time points where observations are recorded */
int nObservationPoints;

/* Maps the position of the variable in the data table to the
   position of the variable in the complete variable list */
int *file_to_var_map;

/* Maps the position of the variable in the complete variable
   list back to its position in the file */
int var_to_file_map[NUM_VARS];

/* The function requires an opened data file that can be read */
void
build_data_table (char *file_name, char numeric_field_names[][NAME_LENGTH])
{
  char c, old_c;
  int i, j;
  int field_length = 0;
  int tmp_field_length = 0;
  FILE *f;

  /* open the file */
  if ((f = fopen (file_name, "rt")) == NULL)
    {
      fprintf (stderr, "File open failed\n");
      exit (-1);
    }


  /* read the first line of the file, counting field names */
  old_c = ' ';
  for (c = fgetc (f); c != '\n'; c = fgetc (f))
    {
      /* assume strict conformance to file specification */
      if (old_c == ' ' && c != ' ')
	{
	  nFields++;
	  tmp_field_length = 1;	/* start a new field */
	}
      else if (old_c != ' ' && c == ' ' && tmp_field_length > field_length)
	{
	  field_length = tmp_field_length;
	}
      else if (old_c != ' ' && c != ' ')
	{
	  tmp_field_length++;
	}
      old_c = c;
    }

  /* count the number of value-filled lines in the data file */
  for (c = fgetc (f); !feof (f); c = fgetc (f))
    {
      /* if the line starts with a space, assume it is blank */
      if (!isspace (c))
	{
	  nObservationPoints++;
	}
      while (c != '\n' && !feof (f))
	{
	  c = fgetc (f);
	}
    }

  /* allocate space for the field names */
  field_names = (char **) malloc (nFields * sizeof (char *));
  ensure_allocation (field_names,
		     "out of memory error while storing field names\n");

  for (i = 0; i < nFields; i++)
    {
      field_names[i] = (char *) malloc (field_length * sizeof (char));
      ensure_allocation (field_names[i],
			 "out of memory error while storing field names\n");
    }

  /* allocate space for the data table */
  observed_values = (double **) malloc (nObservationPoints * sizeof (double *));
  ensure_allocation (observed_values,
		     "out of memory error while storing observed values");

  for (i = 0; i < nObservationPoints; i++)
    {
      observed_values[i] =
	(double *) malloc (nFields * sizeof (double));
      ensure_allocation (observed_values[i],
			 "out of memory error while storing observed values");
    }

  /* read the field names into the array of field names */
  rewind (f);			/* jump back to the beginning */
  /* prepare the maps */
  file_to_var_map = (int *) malloc (nFields * sizeof (int));
  ensure_allocation (file_to_var_map,
		     "out of memory error while storing observed values");

  /* set the map to the default not found value */
  for (i = 0; i < NUM_VARS; i++) {
  	var_to_file_map[i] = -1;
  }
  for (i = 0; i < nFields; i++)
    {
      read_field_name (f, field_names[i]);
      file_to_var_map[i] = -1;	/* default not found */

      /* Map the field to the known variable names if possible */
      for (j = 0; j < NUM_VARS; j++)
	{
	  if (strcmp (numeric_field_names[j], field_names[i]) == 0)
	    {
	      file_to_var_map[i] = j;
	      var_to_file_map[j] = i;
	      printf ("\t\t%s in var[%d]\n", field_names[i], j);
	    }
	}
    }
    /* devour the rest of the names line */
	for(c = fgetc(f); c != '\n'; c = fgetc(f));

  /* read the data into the data table */
  for (i = 0; i < nObservationPoints; i++)
    {
      read_observation_vector (f, observed_values[i]);
    }

  /* close the file */
  fclose (f);
}

/* Extracts a single field name delimited with a ' ' or '\n' 
   Naively assumes that there's enough room in a to store the name
*/
void
read_field_name (FILE * f, char *a)
{
  char c;
  int i = 0;
  for (c = fgetc (f); c != ' ' && c != '\n' && !feof (f); c = fgetc (f))
    {
      a[i++] = c;
    }
  a[i] = '\0';
  /* we want to stay on the line for predictable behavior */
  if (c == '\n')
  	ungetc(c,f);
}

/* Extracts all values on a line of the data file and stores them in a
   Assumes that a is big enough
*/
void
read_observation_vector (FILE * f, double *a)
{
  char c;
  int i = 0;
  int bReadValue = 0;

  for (c = fgetc (f); c != '\n' && !feof (f); c = fgetc (f))
    {
      /* if we see a number, grab it */
      if (c != ' ')
	{
	  ungetc (c, f);
	  fscanf (f, "%lf", &a[i++]);
	  bReadValue = 1;
	}
      else
	{
	  /* if we see a space and we just read a number, get ready 
	     for the next number */
	  if (bReadValue)
	    {
	      bReadValue = 0;
	    }
	  else
	    {
	      /* We've seen two spaces in a row, so a field had an 
	         empty value.  Set this to NAN for easy detection.
	         This keeps us from needing a value for every 
	         observable at every time step */
	      a[i++] = NAN;
	      bReadValue = 1;
	    }
	}
    }
}

void
ensure_allocation (void *a, char *msg)
{
  if (a == NULL)
    {
      fprintf (stderr, "%s", msg);
      exit (-1);
    }
}


/* This function is inefficient right now, but it gets the 
 * right answer, so don't complain too much. */
double get_most_recent_observation(int column, double timestep) {
	int i;
	int old_row = 0;
	/* ASSUME: Time is stored in the first column */
	/* while we have more rows and we haven't found a row from a time
	 * in the future, keep looking */
	for (i = 0; i < nObservationPoints && observed_values[i][0] <= timestep; i++) {
		/* if there's an observed value, go ahead and update */
		if (!isnan(observed_values[i][column])) {
			old_row = i;
		}
	}
	return observed_values[old_row][column];
}

double
get_current_observation(int column, double timestep, double timestep_size) {
	int i = 0;
	/* observed_values[i][0] is the time of the ith set of observations */
	while (i < nObservationPoints && observed_values[i][0] < timestep) {
		i++;
	}
	
	/* if i is out of bounds, we better return NAN */
	/* the timestep needs to be evaluated with respect to some tolerance */
	/* since we're dealing w/ doubles, we can't do a strict equality check
	 * without crossing our fingers.  Instead we look at the TIMESTEP size
	 * and see whether the current timestep falls within a particular range,
	 * then we'll accept it.
	 */
	 /* we make our range much wider than it really needs to be, but wide enough
	  * that reasonable (?) situations won't cause any problems */
	 
	if (i >= nObservationPoints ||
		(((observed_values[i][0] - 0.5 * timestep_size) < timestep) &&
		 (timestep < (observed_values[i][0] + 0.5 * timestep_size))))
		return observed_values[i][column];
	else
		return NAN;		
}

void
print_data_tables ()
{
  int i, j;
  for (i = 0; i < nFields; i++)
    {
      printf ("%s\t", field_names[i]);
    }
  printf ("\n");

  for (i = 0; i < nObservationPoints; i++)
    {
      for (j = 0; j < nFields; j++)
	{
	  if (!isnan (observed_values[i][j]))
	    {
	      printf ("%f\t", observed_values[i][j]);
	    }
	  else
	    {
	      printf ("unobserved\t");
	    }
	}
      printf ("\n");
    }
}
