/*
 * simulator.c: main code for simulation. A c package cvode is used for 
 *              solving differential equations.
 *
 * There are 3 modes for simulation: SEARCH-MODE, PREDICT_MODE, DIAGONOSE_MODE
 * SEARCH_MODE: I don't understand (Mei).
 * PREDICT_MODE: used for prediction. In this mode, when we have a predicted and an observed value for a variable, 
 *               we always will use the PREDICTED value for further computation. 
 * DIAGNOSE_MODE: used for diagnosing. In this mode, when we have a predicted and an observed value for a variable, 
 *                we will use the OBSERVED value for further computation (prediction) . 
 * 
 * Author(s): Javier Nicolas Sanchez
 *            Mei Marker
 */

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "ODE.h"
#include "util.h"
#include "data.h"
#include "output.h"		/* write simulation results to files */

//Time, is the internal clock of the model
double t;
int START_TIME;
int NUM_TIMESTEPS;
double TIMESTEP;

#include "model.h"

/* makes sure that all exogenous values are placed in the variables
 * for the next time through the set of equations
 */
void fill_exogenous_variables() {
	int i;
	for (i = 0; i < NUM_VARS; i++)
	{
		if (exogenous[i]) {
			old_var[i] = var[i];
      		var[i] = get_most_recent_observation(var_to_file_map[i], t);
		}
	}
}

void
simulate_one_timestep (int mode)
{
  int i;

  /* reset the state so that no processes are reportedly evaluated */
  for (i = 0; i < NUM_PROCS; i++)
    evaluated_proc[i] = 0;

#ifdef INSTANTANEOUS
  fill_exogenous_variables();
  evalVars (mode);
  for (i = 0; i < NUM_PROCS - 1; i++)
    {
      if ((*test_proc[i]) () == 1)
	{
	  evaluated_proc[i] = 1;
	}
    }
#else

  /* Ensure that all ODEs are processed first */
  if (evaluated_proc[NUM_PROCS - 1] == 0
      && (*test_proc[NUM_PROCS - 1]) () == 1)
    {
      (*proc[NUM_PROCS - 1]) ();
      evaluated_proc[NUM_PROCS - 1] = 1;
    }

  /* Process the algebraic equations */
  fill_exogenous_variables();
  evalVars (mode);

  /* Report that we have evaluated the processes */
  if (mode != DIAGNOSE_MODE)
    {
      for (i = 0; i < NUM_PROCS - 1; i++)
	{
	  if ((*test_proc[i]) () == 1)
	    {
	      evaluated_proc[i] = 1; 
	    }
	}
    }
#endif
}


#ifndef INSTANTANEOUS

/*y vector_has the initial value.*/
void
simulate_diff_eqs (int NEQ, float *values,
		   void (*f) (integer N, real t, N_Vector y, N_Vector ydot,
			      void *f_data))
{
  void *cvode_mem;
  N_Vector y;
  real auxT;
  int flag, i;

  //printf("timestep = %f\n",TIMESTEP);
  /* ********************************************************** */
  //Do the initialization according to the initial value
  y = N_VNew (NEQ, NULL);
  for (i = 0; i < NEQ; i++)
    N_VIth (y, i) = values[i];
  /* ********************************************************** */

  cvode_mem =
    CVodeMalloc (NEQ, f, 0, y, ADAMS, NEWTON, SS, &reltol, &abstol, NULL,
		 NULL, FALSE, NULL, NULL, NULL);
  if (cvode_mem == NULL)
    {
      printf ("CVodeMalloc failed.");
      exit (-1);
    }
  CVDiag (cvode_mem);

  /* ********************************************************** */
  //This code also depends on what we want from the simulation
  if (t > 0)
    {
      for (i = 0; i < NEQ; i++)
	printf ("values[%d] = %f\n", i, values[i]);
      printf ("values: %f,%f,%f\n", values[1], values[3], values[4]);
      flag = CVode (cvode_mem, TIMESTEP, y, &auxT, NORMAL);
      if (flag != SUCCESS)
	{
	  printf ("CVode returned error flag = %d\n", flag);
	  exit (-1);
	}
      for (i = 0; i < NEQ; i++)
	values[i] = N_VIth (y, i);
      for (i = 0; i < NEQ; i++)
	printf ("values[%d] = %f\n", i, values[i]);
      printf ("values: %f,%f,%f\n", values[1], values[3], values[4]);
    }
  /* ******************************************************** */

  CVodeFree (cvode_mem);
  N_VFree (y);
}


#endif
