/*
 * simulator_main.c: Main functions for simulation. 
 *
 * Author(s): Javier Nicolas Sanchez
 *            Mei Marker
 *
 */

#include "simulator.h"
#include "model.h"
#include "data.h"
#include "util.h"
#include "output.h"
#include <stdio.h>
#include <stdlib.h>

void test_print_variable_values();

int
main (int argc, char *argv[])
{
  int i;
  /* for models w/ only algebraic equations, this will count the number of
   * data seen so far.  this differs from the time step, which could start
   * somewhere other than 0.
   */
  int data_counter = 0;
  FILE *outStream;

  /* parse the command line */
  /* instantaneous models only require a data file. */
  /* ODE models must also be given a number of time steps to execute */
#ifdef INSTANTANEOUS
  if (argc == 3) {
	TIMESTEP = 1;
  }
#else
  if (argc == 5)
    {
      NUM_TIMESTEPS = atoi (argv[3]);
      TIMESTEP = atof (argv[4]);
    }
#endif
  else
    {
      EXIT ("Incorrect number of command line arguments");
    }

  initialize_model ();
  printf ("model initialized\n");

  build_data_table (argv[1], numeric_field_names);
  printf ("observable data module initialized\n");

  /* prepare  the output stream */
  outStream = fopen (argv[2], "wt");
  print_full_result_header (outStream);

  /* Set the starting value of all variables to their observed value */
  /* If the observable variables have been seeded give the specified value, 
     otherwise the variable gets the value 0 */

  /* ASSUME that the first m members of var[ ] are the same as the m members
     of numeric_field_names[ ] -- check that this is enforced. */
  /* zero out all variables */
  for (i = 0; i < NUM_VARS; i++)
    {
      old_var[i] = 0;
      var[i] = 0;
      result[i] = 0;
    }

  /* fill in the observed values at time 0 */
  for (i = 0; i < NUM_VARS; i++)
    {
      old_var[i] =
	  var_to_file_map[i] >= 0 ? observed_values[0][var_to_file_map[i]] : 0;
      var[i] = old_var[i];
      result[i] = old_var[i];
    }

  /* ASSUME: time is the first field and that you want to match the start
   * time of your simulation with the start time in your input file */
   t = START_TIME = observed_values[0][0];

  /* Evaluate any algebraic equations */
   evalVars(PREDICT_MODE);

   /* Make sure to prime the system with these results */
   for (i = 0; i < NUM_VARS; i++)
     {
        result[i] = old_var[i] = var[i];
     }
  printf ("simulation initialized\n");

  /* output the values of the initial observations */
  /* assumes intrinsic values are at the end of the arrays */
  for (i = 0; i < NUM_NUMERIC_INPUTS + NUM_OUTPUTS; i++)
    {
      printf ("%s: var[%d] = %f\n", numeric_field_names[i], i, var[i]);
    }

  /* output the initial conditions and update t */
  print_full_output (outStream);
  t+=TIMESTEP;
  data_counter++;
  
  /* Main loop for the simulation */
#ifdef INSTANTANEOUS
  while (data_counter < nObservationPoints)
#else
  while (t < (NUM_TIMESTEPS*TIMESTEP) + START_TIME)
#endif
    {
      /* Do the simulation for the current input vector */
      /* We are simulating, so use PREDICT_MODE.  This ensures that the variables (v[i]) will be updated */
      simulate_one_timestep (PREDICT_MODE);

      for (i = 0; i < NUM_PROCS - 1; i++)
		{
	  		if ((*test_proc[i]) () == 1)
	    		evaluated_proc[i] = 1;
		}

      /* Put the values from this cycle into result so that they can be printed
         correctly by printOuputStream */
      /* BUG values may be in observed and not in var */
      for (i = 0; i < NUM_VARS; i++)
		result[i] = var[i];

      /* Output the results of simulating with the current input vector */
      /* This means that any exogenous variables have their values from the 
       * previous time step where recorded values were provided */
      print_full_output (outStream);

      /* keep track of the t-1 values */
      /* BUG values may be in observed and not in var */
      for (i = 0; i < NUM_NUMERIC_INPUTS + NUM_OUTPUTS; i++)
		old_var[i] = var[i];

	   /* test_print_variable_values(); */ 

       
	  /* ASSUME that the first m members of var[ ] are the same as the m members
         of numeric_field_names[ ] -- check that this is enforced. */
	  
      /* update the time */
      t += TIMESTEP;
	  data_counter++;
    }

  return 0;
}



 /* debugging code that prints the values at each time step to the screen */
void test_print_variable_values() {
	int i;
	for (i = 0; i < NUM_VARS; i++)
	{
		printf("observed[%d] = %f\n", i, observed[i]);
		printf ("var[%d] = %f\n", i, var[i]);
	}
}
